/*********************************************************************/
/* neblokeys.c (v. 0.2)                                              */
/* Asymmetric login key generator for neblo			     */
/* Copyright (C) 2006 - Paolo Gianrossi <paolino@member.fsf.org>     */
/* 								     */
/* This program is free software; you can redistribute it and/or     */
/* modify it under the terms of the GNU General Public License as    */
/* published by the Free Software Foundation; either version 2 of    */
/* the License, or (at your option) any later version.		     */
/*								     */
/* This program is distributed in the hope that it will be useful,   */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of    */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU */
/* General Public License for more details.			     */
/*								     */
/* You should have received a copy of the GNU General Public License */
/* along with this program; if not, write to the Free Software	     */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA	     */
/* 02111-1307, USA.						     */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <neblo_secret.h>
#include <argp.h>
#include <config.h>
#include <openssl/rsa.h>
#include <openssl/rand.h>

#include <time.h>

const char *argp_program_version =
"neblokeys (" PACKAGE_STRING ")";
const char *argp_program_bug_address =PACKAGE_BUGREPORT;

/* Program documentation. */
static char doc[] =
"neblokeys -- Make and test neblo asymmetric keys";

/* A description of the arguments we accept. */
static char args_doc[] = "";

/* The options we understand. */
static struct argp_option options[] = {
  {"test", 't', "MESSAGE", OPTION_ARG_OPTIONAL, "Instead of generating keys, test existing keys. "
                                                "MESSAGE will be the string used for the test."},
   {"verbose",  'v', 0,      0,  "Produce verbose output (to stderr)" },
  {"stdout",   's', 0,      0,  "Output to standard output instead of files"},
  {"bits",     'b', "LENGTH",      0,  "Make a key LENGHT bits long"},
  {"exponent", 'e', "EXP", 0, "Set key exponent to EXP"},
  {"file",   'f', "PREFIX", 0,
   "Output to PREFIX.sk and PREFIX.pk instead of default files (neblo.[sk|pk]),"
   " or (with the `test' option), test keys PREFIX.sk and PREFIX.pk"},
  { 0 }
};

#define VERBOSE_PRINT(flag, ...) if(flag){fprintf(stderr, __VA_ARGS__);}
/* Used by `main' to communicate with `parse_opt'. */
struct arguments
{
  char *prefix;                /* ARG1 & ARG2 */
  int verbose;
  int stdout;
  int bits;
  int test;
  char *message;
  int exp;
};

/* Parse a single option. */
static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  /* Get the INPUT argument from `argp_parse', which we
     know is a pointer to our arguments structure. */
  struct arguments *arguments = state->input;

  switch (key)
    {
    case 't':
      arguments->test = 1;
      arguments->message=arg;

    case 'v':
      arguments->verbose = 1;
      break;
    case 'f':
      arguments->prefix = arg;
      break;
    case 's':
      arguments->stdout=1;
      break;
    case 'e':
      arguments->exp=atoi(arg);
      break;
    case 'b':
      arguments->bits=atoi(arg);
      break;
    case ARGP_KEY_ARG:
      	argp_usage (state);
	break;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

/* Our argp parser. */
static struct argp argp = { options, parse_opt, args_doc, doc };

int main (int argc, char **argv)
{
  struct arguments arguments;
  char *pubfile;
  char *secfile;
  register int retry=3;
  int v=0;
  FILE *pf, *sf;
  RSA *private=RSA_new();
  RSA *public=RSA_new();

  /* Default values. */
  arguments.stdout = 0;
  arguments.verbose = 0;
  arguments.prefix = "neblo";
  arguments.bits=DEFAULTBITKEYSIZE;
  arguments.exp=DEFAULTKEYEXP;
  arguments.message="This is the default test message.";
  arguments.test=0;

  /* Parse our arguments; every option seen by `parse_opt' will
     be reflected in `arguments'. */
  argp_parse (&argp, argc, argv, 0, 0, &arguments);
  v=arguments.verbose;

  if(!arguments.test){
    if(arguments.stdout){
      VERBOSE_PRINT(v, "Keys will be output to standard output.\n");
      pf=stdout;
      sf=pf;
    }
    else
      {
	pubfile=(char *)malloc((strlen(arguments.prefix) + 5) * sizeof(char));
	secfile=(char *)malloc((strlen(arguments.prefix) + 5) * sizeof(char));
	strcpy(pubfile, arguments.prefix);
	strcat(pubfile, ".pk");
	strcpy(secfile, arguments.prefix);
	strcat(secfile, ".sk");
	VERBOSE_PRINT(v, "Public key and secret key will be output respectively to files: `%s' and `%s'\n",
		      pubfile, secfile);
      }

    srand(time(NULL));
    while (retry > 0)
      {
	VERBOSE_PRINT(v, "Generating key pair"
		      " of %d bits. Exp is %d... ", arguments.bits, arguments.exp);

	private=RSA_generate_key(arguments.bits, arguments.exp, NULL, NULL);

	VERBOSE_PRINT(v, "done.\nChecking key validity... ");

	if(!RSA_check_key(private)){
	  if(retry >1){
	    VERBOSE_PRINT(v, "FAILED! Not saving keys! Retrying...\n");
	    retry--;
	    continue;
	  }
	  else {
	    fprintf(stderr, "%s: Fatal error: Keys generated were invalid for %d consecutive times...\n"
		    "Aborting.\n", argp_program_version, retry);
	    exit(EXIT_FAILURE);
	  }
	}
	else{
	  retry = 0;
	}
      }

    if(!arguments.stdout){
      VERBOSE_PRINT(v, "done.\nSaving private key to %s... ", secfile);
      sf=fopen(secfile, "w");
      if(!sf){
	perror("Opening secret file");
	exit(EXIT_FAILURE);
      }
    } else {
      VERBOSE_PRINT(v, "done.\nWriting private key to standard output...");
    }
    /* BIGNUM *n;              // public modulus */
    /* BIGNUM *e;              // public exponent */
    /* BIGNUM *d;              // private exponent */
    /* BIGNUM *p;              // secret prime factor */
    /* BIGNUM *q;              // secret prime factor */
    /* BIGNUM *dmp1;           // d mod (p-1) */
    /* BIGNUM *dmq1;           // d mod (q-1) */
    /* BIGNUM *iqmp;           // q^-1 mod p */
    if(arguments.stdout)
      fprintf(sf, "\n---Private Key Begin--\n");

    fprintf(sf, "%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s",
	    BN_bn2hex(private->n),
	    BN_bn2hex(private->e),
	    BN_bn2hex(private->d),
	    BN_bn2hex(private->p),
	    BN_bn2hex(private->q),
	    BN_bn2hex(private->dmp1),
	    BN_bn2hex(private->dmq1),
	    BN_bn2hex(private->iqmp));
    if(arguments.stdout)
      fprintf(sf, "\n---Private Key End--\n");

    if(!arguments.stdout)
      fclose(sf);
    VERBOSE_PRINT(v, "done\n");

    if(!arguments.stdout){
      VERBOSE_PRINT(v, "Saving public key to %s... ", pubfile);
      pf=fopen(pubfile, "w");
      if(!pf){
	perror("Opening public file");
	exit(EXIT_FAILURE);
      }
    } else {
      VERBOSE_PRINT(v, "Writing public key to standard output...");
    }
    public->n=private->n;
    public->e=private->e;
    if(arguments.stdout)
      fprintf(pf, "\n---Public Key Begin--\n");
    fprintf(pf, "%s\n%s\n",BN_bn2hex(public->n),BN_bn2hex(public->e));
    if(arguments.stdout)
      fprintf(pf, "\n---Public Key End--\n");

    if(!arguments.stdout){
      fclose(pf);
      fprintf(stderr, "Done\n--Public Key Begin--\n");
      RSA_print_fp(stdout, public, 0);
      fprintf(stderr, "--Public Key End--\n");
    }
    free(pubfile);
    free(secfile);
    return 0;
  }
  else
  /* TESTING */
    {
        unsigned char *cipher;
	unsigned char *clearMsg;
	int long_msg;

	char fieldStr[MAXBYTEKEYSIZE-1];
	pubfile=(char*)malloc((strlen(arguments.prefix)+5)*sizeof(char));
	secfile=(char*)malloc((strlen(arguments.prefix)+5)*sizeof(char));
	strcpy(pubfile,arguments.prefix);
	strcat(pubfile,".pk");
	strcpy(secfile,arguments.prefix);
	strcat(secfile,".sk");
	VERBOSE_PRINT("Testing keypair %s - %s\n", pubfile, secfile);
	pf=fopen(pubfile, "r");	/* public key file */
	if(!pf){
	  perror("reading the public file");
	  exit(EXIT_FAILURE);
	}
	sf=fopen(secfile, "r");	/* secret key file */
	if(!sf){
	  perror("reading the secret file");
	  exit(EXIT_FAILURE);
	}

	_get_public_key_(pf, public, fieldStr);

	if(arguments.verbose){
	  fprintf(stderr, "--Public Key Begin--\n");
	  RSA_print_fp(stderr, public, 0);
	  fprintf(stderr, "--Public Key End--\n\n");
	}

	_get_private_key_(sf, private, fieldStr);

	/* Encriptamos el msg */
	VERBOSE_PRINT(v,"Encoding message...");
	cipher= (unsigned char *) malloc (RSA_size(public));
	long_msg= (strlen(arguments.message) +1)*sizeof(char);
	RSA_public_encrypt(long_msg, arguments.message, cipher, public, \
			   RSA_PKCS1_OAEP_PADDING);
	VERBOSE_PRINT(v,"done.\nDecoding message...");
	/* decrypt msg */
	clearMsg= (unsigned char *) malloc(RSA_size(private));
	RSA_private_decrypt(RSA_size(private), cipher, clearMsg,private, \
			    RSA_PKCS1_OAEP_PADDING);
	VERBOSE_PRINT(v, "done.\n");

	VERBOSE_PRINT(v, "The original message: `%s'\n", arguments.message);
	VERBOSE_PRINT(v, "After encoding and decoding back: `%s'\n", clearMsg);
	if(!strcmp(arguments.message, clearMsg)){
	  fprintf(stdout, "Test succeded. Keypair [%s - %s] is working.\n", pubfile, secfile);
	}
	else {
	  fprintf(stdout, "Failure! Keypair [%s - %s] is broken.\n", pubfile, secfile);
	}

	RSA_free(public);
	RSA_free(private);
	free(cipher);
	free(clearMsg);
	free(pubfile);
	free(secfile);


	return 0;


    }
  /* should never be able to get here */
  return 1;

}
