/*********************************************************************/
/* miscell.c                                                         */
/* miscellaneous support routines      				     */
/* see COPYING in parent directory                                   */
/*********************************************************************/

#include <neblo_.h>
#include <queues.h>
#include <neblo_macros.h>
#include <neblo_secret.h>


/*  Ciphred communication  */



void
_close_
( csocket_t **cs )
{
  if (unlikely((*cs) == NULL))
    return;
  close((*cs)->sock);
  pthread_mutex_lock(&malloc_lock);
  free((*cs)->ses_k);
  free((*cs)->bf_ses_k);
  free((*cs)->msg);
  free((*cs)->combuf);
  free((*cs));
  *cs = NULL;
  pthread_mutex_unlock(&malloc_lock);
  return;
}



void
_openout_
(
  struct in_addr hostaddr, u_int16_t port, csocket_t **cs, RSA *pub_k, int *rc
)
{
  csocket_t *csock;
  unsigned int i;
  int msg_size, r;
  struct sockaddr_in SA;

  if (pub_k == NULL) {
    S(fprintf(stderr,"_openout_():  NULL pubkey\n"));
    *rc = -3;
    *cs = NULL;
    return;
  }

  _do_(_calloc_(csock,sizeof(csocket_t)),!csock,"_openout_: csock calloc");

  _do_(_calloc_(csock->ses_k,SESSIONKEYSIZE),
       !csock->ses_k, "_openout_: ses_k calloc");
  _do_(_calloc_(csock->bf_ses_k,sizeof(BF_KEY)),
       !csock->bf_ses_k, "_openout_: bf_ses_k calloc");
  for (i = 0; i < 8; i++)
    csock->ivec[i] = 0;
  _do_(_calloc_(csock->msg,RSA_size(pub_k)+BUFLEN+SLACK),
       !csock->msg, "_openout_: msg calloc");
  _do_(_calloc_(csock->combuf,COMBUF_SIZE),
       !csock->combuf, "_openout_: combuf calloc");

  memset(&SA, 0, sizeof(SA));
  SA.sin_family = AF_INET;
  SA.sin_port = htons(port);
  SA.sin_addr = hostaddr;
  _do_((csock->sock) = socket(AF_INET,SOCK_STREAM,0),(csock->sock)<0,
       "_openout_: socket");

  r = connect(csock->sock,(struct sockaddr *)&SA,sizeof(SA));
  if (unlikely(r < 0)) {
    perror("_openout_: connect");
    _close_(&csock);
    *rc = -2;
    goto exit;
  }

  /* choose a session key */
  for (i=0;i<SESSIONKEYSIZE>>2;i++)
    (csock->ses_k)[i] =
    (unsigned long)(rand()*(double)(1<<31)/(double)RAND_MAX);
  strncpy(csock->msg,"howdy ",strlen("howdy "));
  /* encrypt the session key by partner public key */
  pthread_mutex_lock(&malloc_lock);
  r = RSA_public_encrypt(SESSIONKEYSIZE,
        (unsigned char *)(csock->ses_k),
        (unsigned char *)&csock->msg[strlen("howdy ")],
        (pub_k),RSA_PKCS1_OAEP_PADDING);
  pthread_mutex_unlock(&malloc_lock);
  if (unlikely(r < 0)) {
    fprintf(stderr,"_openout_: couldn't encrypt session key\n");
    _close_(&csock);
    *rc = -1;
    goto exit;
  }
  msg_size = RSA_size(pub_k) + strlen("howdy ");

  /* send session key to partner, ciphred by its own pub key */
  r = write(csock->sock,csock->msg,msg_size);
  if (unlikely(r <= 0)) {
    _close_(&csock);
    *rc = -1;
    goto exit;
  }

  /* initialize the Blowfish simmetric cipher with session key */
  pthread_mutex_lock(&malloc_lock);
  BF_set_key(csock->bf_ses_k,SESSIONKEYSIZE,csock->ses_k);
  pthread_mutex_unlock(&malloc_lock);

  *rc = 0;

exit:

  /* quitting... */
  *cs = csock;
  return;
}



void
_openin_
( u_int32_t saddr, u_int16_t *pt, csocket_t **cs, int *rc )
{
  csocket_t *csock;
  int i;
  struct sockaddr_in SA;
  u_int16_t port;

  _do_(_calloc_(csock,sizeof(csocket_t)),!csock, "_openin_: csock calloc");
  /* these are useless... */
  _do_(_calloc_(csock->ses_k,SESSIONKEYSIZE),
       !csock->ses_k, "_openin_: ses_k calloc");
  _do_(_calloc_(csock->bf_ses_k,sizeof(BF_KEY)),
       !csock->bf_ses_k, "_openin_: bf_ses_k calloc");
  for (i = 0; i < 8; i++)
    csock->ivec[i] = 0;
  _do_(_calloc_(csock->combuf,COMBUF_SIZE),
       !csock->combuf, "_openin_: combuf calloc");

  port = *pt;
  memset(&SA, 0, sizeof(SA));
  SA.sin_family = AF_INET;
  SA.sin_addr.s_addr = saddr;

  for (;;) {
    SA.sin_port = htons(port);
    _do_((csock->sock) = socket(AF_INET,SOCK_STREAM,0),(csock->sock)<0,
         "_openin_: socket");
    *rc = bind(csock->sock,(struct sockaddr *)&SA,sizeof(SA));
    if (likely(!*rc)) break;
    if ((*rc) && errno == EADDRINUSE) {
      /* port already taken.  go to next one... */
      close(csock->sock);
      fprintf(stderr,"port %u not available. trying %u...\n",
              port,port+100);
      port += 100;
      continue;
    }
    /* any other err on bind is fatal */
    perror("_openin_: bind"); exit(1);
  }

  *pt = port;
  *cs = csock;
  *rc = 0;
  return;
}



void
_accept_
(
  csocket_t *old_cs, struct sockaddr *psa_ptr, socklen_t *size_psa,
  csocket_t **cs, int *rc
)
{

  int i;


  _do_(_calloc_(*cs,sizeof(csocket_t)),!*cs,"_openin_: *cs calloc");
  _do_(_calloc_((*cs)->ses_k,SESSIONKEYSIZE),
       !(*cs)->ses_k, "_openin_: ses_k calloc");
  _do_(_calloc_((*cs)->bf_ses_k,sizeof(BF_KEY)),
       !(*cs)->bf_ses_k, "_openin_: bf_ses_k calloc");
  for (i = 0; i < 8; i++)
    (*cs)->ivec[i] = 0;
  _do_(_calloc_((*cs)->combuf,COMBUF_SIZE),!(*cs)->combuf,
       "_openin_: combuf calloc");
  (*cs)->sock = accept(old_cs->sock,psa_ptr,size_psa);
  if ((*cs)->sock < 0)
    *rc = -1;
  else
    *rc = 0;
  return;

}



void
_establishin_
(
  csocket_t *csock, RSA *pub_k, RSA *seckey, FILE *skf, char *seckeybuf,
  int *rc
)
{

  int nbytes,msg_size;
  char *tmp;
  int r;


  _do_(_calloc_(csock->msg,RSA_size(pub_k)+BUFLEN+SLACK),!csock->msg,
       "_establishin_: msg calloc");
  _do_(_calloc_(tmp,RSA_size(pub_k) + BUFLEN + SLACK),!tmp,
       "_establishin_: tmp calloc");

  /* get session key, encrypted by this peer's public key */
  msg_size = RSA_size(pub_k) + strlen("howdy ");
  nbytes = 0;
  for (;;) {
    r = read(csock->sock,csock->msg+nbytes,msg_size-nbytes);
    /* if conn. closed or something went wrong... */
    if (unlikely(r < 0)) break;
    nbytes += r;
    if (csock->msg[nbytes-1] == '\n') break;
    if (nbytes >= msg_size) break;
  }
  if (unlikely(r <= 0)) {
    *rc = -1;
    return;
  }
  if (nbytes)
    csock->msg[nbytes] = 0;
  if (unlikely(strncmp(csock->msg,"howdy ",strlen("howdy ")))) {
    *rc = -1;
    return;
  }

  /* decrypt session key by using this peer's private key */
  pthread_mutex_lock(&malloc_lock);
  _get_private_key_(skf,seckey,seckeybuf);
  r = RSA_private_decrypt(RSA_size(pub_k),
                          (unsigned char *)&csock->msg[6],
                          (unsigned char *)(tmp), seckey,
                          RSA_PKCS1_OAEP_PADDING);
  _clean_private_key_(seckey,seckeybuf,KEYBUFLEN+SLACK);
  memcpy(csock->ses_k,tmp,SESSIONKEYSIZE);
  free(tmp);
  pthread_mutex_unlock(&malloc_lock);
  if (unlikely(r < 0)) {
    fprintf(stderr,"_establishin_: couldn't decrypt session key\n");\
    *rc = -1;
    return;
  }

  /* initialize the Blowfish simmetric cipher with session key */
  pthread_mutex_lock(&malloc_lock);
  BF_set_key(csock->bf_ses_k,SESSIONKEYSIZE,csock->ses_k);
  pthread_mutex_unlock(&malloc_lock);

  *rc = 0;
  return;

}



void
__read__
( csocket_t *csock, void *buf, unsigned int maxlen, int *rc, long timeout )
{

  fd_set rfd;
  int dsize;
  unsigned int nbytes;
  struct timeval tout;
  int r;


  if (unlikely(csock == NULL)) {
    *rc = -100;
    return;
  }

  FD_ZERO(&rfd);
  FD_SET((csock)->sock,&rfd);
  dsize = getdtablesize();\

  /* get msg size.  lost conns will fail upon timeout expiration */\
  nbytes = 0;
  for (;;) {
    if ((timeout) >= 0) {
      tout.tv_sec = (timeout);
      tout.tv_usec = 0;
      r = select(dsize,&rfd,NULL,NULL,(struct timeval *)&tout);
    }
    else
      r = select(dsize,&rfd,NULL,NULL,(struct timeval *)NULL);
    if (unlikely(r <= 0))
      break; /* err or time out */
    r = read(csock->sock,(char *)&csock->combuf_len+nbytes,
             sizeof(csock->combuf_len)-nbytes);
    /* if conn. closed or something went wrong... */
    if (unlikely(r <= 0))
      break; /* conn. closed or something went wrong */
    nbytes += r;
    if (nbytes >= sizeof(csock->combuf_len)) break;
  }
  if (unlikely(r <= 0)) {
#ifdef DEBUG
    perror("__read__(): reading msg len");
#endif
    *rc = -200;
    return;
  }
  if (unlikely(csock->combuf_len > maxlen)) {
    *rc = -300;
    return;
  }
  if (unlikely(csock->combuf_len > COMBUF_SIZE)) {
    *rc = -400;
    return;
  }

  /* get the actual msg.  lost conns will fail upon timeout expiration */
  nbytes = 0;
  for (;;) {
    if ((timeout) >= 0) {
      tout.tv_sec = (timeout);
      tout.tv_usec = 0;
      r = select(dsize,&rfd,NULL,NULL,(struct timeval *)&tout);
    }
    else
      r = select(dsize,&rfd,NULL,NULL,(struct timeval *)NULL);
    if (r <= 0)
      break; /* err or timed out */
    r = read(csock->sock,csock->combuf+nbytes,csock->combuf_len-nbytes);
    if (unlikely(r <= 0))
      break; /* conn. closed or something went wrong */
    nbytes += r;
    if (nbytes >= csock->combuf_len) break;
  }
  if (unlikely(r <= 0)) {
#ifdef DEBUG
    perror("__read__(): reading msg");
#endif
    *rc = -500;
    return;
  }

  /* decode msg, from combuf into buf.  Blowfish preserves length */
  pthread_mutex_lock(&malloc_lock);
  csock->num = 0;
  BF_cfb64_encrypt((unsigned char *)csock->combuf,(unsigned char *)buf,
                   (long)csock->combuf_len,csock->bf_ses_k,csock->ivec,
                   &csock->num,BF_DECRYPT);
  pthread_mutex_unlock(&malloc_lock);

  rc = 0;
  return;

}



void
__write__
( csocket_t *csock, void *buf, unsigned int len, int *rc )
{

  int r;


  if (unlikely(csock == NULL)) {
    *rc = -1; /* yet unsure whether peer is DEAD or not yet connected */
    return;
  }

  /* encode msg, from buf into combuf. Blowfish preserves data length */
  csock->combuf_len = len;
  pthread_mutex_lock(&malloc_lock);
  csock->num = 0;
  BF_cfb64_encrypt((unsigned char *)buf,(unsigned char *)csock->combuf,
                   (long)len,csock->bf_ses_k,csock->ivec,
                   &csock->num,BF_ENCRYPT);
  pthread_mutex_unlock(&malloc_lock);

  /* send out, prepended by msg size */
  r = write(csock->sock,&csock->combuf_len,sizeof(csock->combuf_len));
  r = write(csock->sock,csock->combuf,csock->combuf_len);
  if (unlikely(r <= 0)) {
    if (errno == 32) /* broken pipe.  DEAD for sure. */
      *rc = -2;
    else
      *rc = -1;
#ifdef DEBUG
    perror("__write__()");
#endif
    return;
  }

  *rc = 0;
  return;

}



/*  Put a challenge on foreign peer about its supposed pubkey.  */
int
_put_challenge_
( csocket_t *cs, char *cmd, RSA *foreign_pubkey )
{

  char   *chall;
  int     rc = 0;

  _do_(_malloc_(chall,SESSIONKEYSIZE),!chall,
       "_put_challenge_(): chall malloc");
  RAND_bytes(chall,SESSIONKEYSIZE);
  S(rc = RSA_public_encrypt(SESSIONKEYSIZE,chall,cmd,foreign_pubkey,
                            RSA_PKCS1_OAEP_PADDING));
  if (rc < 0) {

#ifdef DEBUG
    fprintf(stderr,"_put_challenge_(): couldn't encrypt challenge\n");
#endif

    goto exit;
  }

  _write_(cs,cmd,RSA_size(foreign_pubkey),rc,/* */);
  _read_(cs,cmd,SESSIONKEYSIZE,rc,TIMEOUT,/* */);
  if (rc < 0) {

#ifdef DEBUG
    fprintf(stderr,"_put_challenge_(): challenge interrupted\n");
#endif

    goto exit;
  }

  if (strncmp(chall,cmd,SESSIONKEYSIZE)) {

#ifdef DEBUG
    fprintf(stderr,"_put_challenge_(): challenge failed\n");
#endif

    rc = -1;
  }

exit:

  _free_(chall);
  return rc;

}



/*  Get a challenge about local pubkey and answer back.  */
int
_get_challenge_
( csocket_t *cs, char *cmd )
{

  char   *chall;
  int     rc = 0;


  _do_(_malloc_(chall,SESSIONKEYSIZE),!chall,
       "_get_challenge_(): chall malloc");

  _read_(cs,cmd,KEYBUFLEN,rc,TIMEOUT,goto exit);

  pthread_mutex_lock(&malloc_lock);
  _get_private_key_(skf,seckey,seckeybuf);
  rc = RSA_private_decrypt(RSA_size(pubkey),cmd,chall,seckey,
                           RSA_PKCS1_OAEP_PADDING);
  _clean_private_key_(seckey,seckeybuf,KEYBUFLEN+SLACK);
  pthread_mutex_unlock(&malloc_lock);

  _write_(cs,chall,SESSIONKEYSIZE,rc,/* */);

exit:

  _free_(chall);
  return rc;

}


/*  Neighbours, fingers, and incoming connections ("friends")  */



void
clean_peer_desc
( peer_t *f )
{

  pthread_mutex_lock(&f->lock);
  cmd_handle_lock(&f->handle);
  _close_(&f->cs);
  if (f->in_cs != NULL)
    _close_(f->in_cs);
  f->status = NONE;
  _RSA_free_(f->pubkey);
  f->deadline = f->from_predecessor = f->from_successor = f->self = 0;
  f->magn = 0;
  cmd_handle_reset(&f->handle);
/*  NOTE:  the ->next ptr is NOT cleaned, to avoid dangling refs  */
  /* delete all pending read msgs coming from peer "f" */
  clean_pend_reads_peer(f);
  cmd_handle_unlock(&f->handle);
  pthread_mutex_unlock(&f->lock);
  return;
  
}



void
clean_peer_desc_nolock
( peer_t *f )
{

  pthread_mutex_lock(&f->lock);
  _close_(&f->cs);
  if (f->in_cs != NULL)
    _close_(f->in_cs);
  f->status = NONE;
  _RSA_free_(f->pubkey);
  f->deadline = f->from_predecessor = f->from_successor = f->self = 0;
  f->magn = 0;
  cmd_handle_reset(&f->handle);
/*  NOTE:  the ->next ptr is NOT cleaned, to avoid dangling refs  */
  /* delete all pending read msgs coming from peer "f" */
  clean_pend_reads_peer(f);
  pthread_mutex_unlock(&f->lock);
  return;
  
}



/*
 * init_friendship():  initiate a communication with another peer.
 *
 * pred_succ: 1 if caller is predecessor of callee, 2 if it is successor,
 *            0 otherwise (common case).
 *
 * it is assumed that the peer descriptor pointed to by parameter "f" be
 * already inserted in a routing table, with status TOCALL or PROBABLYDEAD.
 *
 * caller should have raised lock over peer descriptor before invoking this
 * routine.
 *
 * on success, the routine will change the peer status to OK.
 *
 * on failure, the routine will NOT change peer status, nor will it
 * delete the descriptor from the routing table.
 *
 * returns -2 if remote peer is dead for sure, -1 on other failures (e.g.
 * timeout on _read_), 0 on success.
 */
static int
init_friendship
( peer_t *f, unsigned int pred_succ, pthread_t th_id )
{
  
  int rc;
  char *cmd;
  unsigned int dummy;


  dummy = th_id;

#ifdef DEBUG_F
  S(fprintf(stderr,"[%lu]\tinit_friendship() with peer %u.%u.%u.%u:%u (status %u)\n",
            th_id,
            ((unsigned char *)&f->addr.s_addr)[0],
            ((unsigned char *)&f->addr.s_addr)[1],
            ((unsigned char *)&f->addr.s_addr)[2],
            ((unsigned char *)&f->addr.s_addr)[3],
            f->port,f->status));
#endif

  if (!(f->status & (TOCALL | PROBABLYDEAD))) {
    /* peer is already connected or DEAD or NONE.  do nothing */

#ifdef DEBUG_F
    S(fprintf(stderr,"[%lu]\tpeer has status %u.  quit...\n",
              th_id,f->status));
#endif

    return 0;
  }

  /* when status == TOCALL or previous attempts have failed... */

  _do_(_malloc_(cmd,KEYBUFLEN+SLACK),!cmd,"init_friendship(): cmd malloc");

  rc = 0;
  if (f->cs == NULL) {

#ifdef DEBUG
    S(fprintf(stderr,"[%lu]\topen csock...\n",th_id));
#endif

    _openout_(f->addr,(u_int16_t)f->port,&f->cs,f->pubkey,&rc);
  }
  if (rc < 0)
    /*
     *  Remote peer refuses connections (rc = -2), or dropped connection down
     *  or has troubles with cipher keys (rc = -1).
     *  In the first of such cases, the caller can assume peer is dead.
     */
    goto abort;

  sprintf(cmd,"friendship %u %u",PORT,pred_succ);

#ifdef DEBUG
  S(fprintf(stderr,"[%lu]\ttalking to peer...\n",th_id));
#endif

  _write_(f->cs,cmd,strlen(cmd)+1,rc,goto abort);

#ifdef DEBUG_F
  S(fprintf(stderr,"[%lu]\t<< %s\n",th_id,cmd));
#endif

  /*  Get challenged about our pubkey  */
  rc = _get_challenge_(f->cs,cmd);
  if (rc < 0) {

#ifdef DEBUG
    fprintf(stderr,"[%lu]\tchallenge interrupted\n",th_id);
#endif

    rc = -1;
    goto abort;
  }

#ifdef DEBUG
  fprintf(stderr,"[%lu]\tchallenge done\n",th_id);
#endif

  _read_(f->cs,cmd,BUFLEN,rc,TIMEOUT, rc = -1; goto abort);

#ifdef DEBUG
  S(fprintf(stderr,"[%lu]\t>> %s\n",th_id,cmd));
#endif

  if (!strncmp(cmd,"ok",strlen("ok"))) {
    f->status = OK;

#ifdef DEBUG_F
    S(fprintf(stderr,"[%lu]\tfriendship established, status %u\n",
              th_id,f->status));
#endif

    rc = 0;
  } else {
    rc = -1;
    goto abort;
  }

  _free_(cmd);
  return rc;

abort:

  _free_(cmd);
  /*  These might be redundant but are not wrong here... */
  _close_(&f->cs);
  if (rc == -2)
    return -2;
  return -1;

}



/*
 * call_peer():  possible first contact with a remote peer.
 *
 * from_pred_succ: 1 if caller is predecessor of callee, 2 if it is successor,
 *                 0 otherwise (common case).
 *
 * This routine might change the status of peer "f" (it will not delete
 * peer from routing table though); thread safety is ensured in doing this.
 *
 * returns -2 if peer is dead (f->status was already, or is set to, DEAD)
 * or non-existent (f->status is NONE);
 * -1 is troubles but unsure it is dead (status is left unchanged),
 * 0 on success (status is set to OK if not already so).
 *
 * If flag == 1, status of dest. peer is also marked as "tmpunavail" on
 * failure.
 */
int
call_peer
( peer_t *f, unsigned int from_pred_succ, int flag, pthread_t th_id )
{

  int rc;


  pthread_mutex_lock(&f->lock);

  if (f->status & DEAD || f->status == NONE) {
    pthread_mutex_unlock(&f->lock);
    return -2;
  }

  if (f->status & (TOCALL | PROBABLYDEAD)) {
    rc = init_friendship(f,from_pred_succ,th_id);
    if (rc == -2) {
      /* peer is dead */
      f->status = DEAD;
      pthread_mutex_unlock(&f->lock);
      return -2;
    }
    if (rc < 0) {
      /* peer has troubles but unsure is dead. status is left unchanged. */
      if (flag)
        f->tmpunavail = 1;
      pthread_mutex_unlock(&f->lock);
      return -1;
    }
  }

  pthread_mutex_unlock(&f->lock);
  return 0;

}

static int
call_peer_nolock
( peer_t *f, unsigned int from_pred_succ, int flag, pthread_t th_id )
{

  int rc;


  if (f->status & DEAD || f->status == NONE) {
    return -2;
  }

  if (f->status & (TOCALL | PROBABLYDEAD)) {
    rc = init_friendship(f,from_pred_succ,th_id);
    if (rc == -2) {
      /* peer is dead */
      f->status = DEAD;
      return -2;
    }
    if (rc < 0) {
      /* peer has troubles but unsure is dead. status is left unchanged. */
      if (flag)
        f->tmpunavail = 1;
      return -1;
    }
  }

  return 0;

}



/*
 * Send a msg to a peer.
 *
 * returns -2 if peer is dead (f->status was already, or is set to, DEAD)
 * or non-existent (f->status is NONE);
 * -1 is troubles but unsure it is dead (status is left unchanged),
 * 0 on success (status is set to OK if not already so).
 *
 * If flag == 1, status of dest. peer is also marked as "tmpunavail" on
 * failure.
 */
int
tell_peer
(
  peer_t *peer, unsigned int from_pred_succ, void *buf, unsigned int len,
  int flag, pthread_t th_id
)
{
  int rc;

  rc = call_peer(peer,from_pred_succ,flag,th_id);
  if (rc < 0)
    return rc;
  pthread_mutex_lock(&peer->lock);
  __write__(peer->cs,buf,len,&rc);
  if (rc < 0) {
    if (flag)
      peer->tmpunavail = 1;
    if (rc == -2)
      peer->status = DEAD;
  }
  pthread_mutex_unlock(&peer->lock);
  return rc;
}

int
tell_peer_nolock
(
  peer_t *peer, unsigned int from_pred_succ, void *buf, unsigned int len,
  int flag, pthread_t th_id
)
{
  int rc;

  rc = call_peer_nolock(peer,from_pred_succ,flag,th_id);
  if (rc < 0)
    return rc;
  __write__(peer->cs,buf,len,&rc);
  if (rc < 0) {
    if (flag)
      peer->tmpunavail = 1;
    if (rc == -2)
      peer->status = DEAD;
  }
  return rc;
}



/*
 *  Get an answer from a peer.
 *  If flag == 1, status of source peer is also marked as "tmpunavail" on
 *  failure.
 */
int
listen_peer
( peer_t *peer, void *buf, unsigned int maxlen, long timeout, int flag )
{
  int rc;

  pthread_mutex_lock(&peer->lock);
  __read__(peer->cs,buf,maxlen,&rc,timeout);
  if (rc < 0)
    if (flag)
      peer->tmpunavail = 1;
  pthread_mutex_unlock(&peer->lock);
  return rc;
}

int
listen_peer_nolock
( peer_t *peer, void *buf, unsigned int maxlen, long timeout, int flag )
{
  int rc;

  __read__(peer->cs,buf,maxlen,&rc,timeout);
  if (rc < 0)
    if (flag)
      peer->tmpunavail = 1;
  return rc;
}


/*  Lists and tables of peer descriptors  */



/*
 *  Append a peer desc. to a list (neighbours, or friends).
 *
 *  Caller must NOT have raised ring->lock
 */
void
append_peer_list
( peer_t **head, peer_t **tail, peer_t *peer )
{
  pthread_mutex_lock(&ring->lock);
  peer->next = NULL;
  if (*head == NULL)
    *head = peer;
  else
    (*tail)->next = peer;
  *tail = peer;
  pthread_mutex_unlock(&ring->lock);
  return;
}



/*
 *  Empties a peer list (neighbours, or friends) and dispose all descriptors.
 *
 *  Caller must NOT have raised ring->lock
 */
void
clean_peer_list
( peer_t **head, peer_t **tail )
{

  peer_t *f;


  pthread_mutex_lock(&ring->lock);
  if (*head == NULL) {
    *tail = NULL;
    pthread_mutex_unlock(&ring->lock);
    return;
  }
  while (*head != NULL) {
    f = *head;
    *head = (*head)->next;
    clean_peer_desc(f);
    _free_(f);
  }
  *tail = NULL;
  pthread_mutex_unlock(&ring->lock);
  return;

}



/*
 *  Purge a peer list (neighbours, or friends) from DEAD entries, and
 *  dispose those descriptors.
 *
 *  Caller must NOT have raised ring->lock
 */
void
purge_peer_list
( peer_t **head, peer_t **tail )
{

  peer_t *prev,*f,*ff;


  pthread_mutex_lock(&ring->lock);
  if (*head == NULL) {
    *tail = NULL;
    pthread_mutex_unlock(&ring->lock);
    return;
  }
  /* move head forward, until first alive peer */
  while (*head != NULL && (*head)->status & DEAD) {
    f = *head;
    *head = (*head)->next;
    clean_peer_desc(f);
    _free_(f);
  }
  if (*head == NULL)
    *tail = NULL;
  else {
    prev = *head;
    for (prev=*head, f=(*head)->next; f!=NULL;) {
      if (!(f->status & DEAD)) {
        prev = f;
        f = f->next;
        continue;
      }
      ff = f;
      f = f->next;
      prev->next = f;
      clean_peer_desc(ff);
      _free_(ff);
    }
    *tail = prev;
  }
  pthread_mutex_unlock(&ring->lock);
  return;

}



/*
 * is_friend(): check if a given incoming request comes through an already
 * established incoming connection (a "friend"), by searching the friend table.
 *
 * Returns ptr to friend if requestor was such, NULL otherwise.
 *
 * Caller must NOT have raised ring->lock
 */
peer_t *
is_friend
( csocket_t *cs )
{

  peer_t *res = NULL;


  if (cs == NULL) return NULL;

  forall_friends(f,
    /*  We are in friend list, so *f->in_cs != NULL for sure.  */
    if (cs == *f->in_cs) {
      res = f;
      break;
    }
  );
  if (!ring->enabled) {
    /*  Ring not yet enabled.  In this case accept only from predecessor */
    if (res != NULL && !res->from_predecessor)
      return NULL;
  }
  return res;

}



/*
 * predecessor(): search in the friend list (incoming conns) for the peer
 * descriptor marked as "predecessor", that is, the peer who welcomed the local
 * peer in the ring.
 *
 * Returns ptr to friend if found and not DEAD, NULL otherwise.
 *
 * Caller must NOT have raised ring->lock
 */
peer_t *
predecessor
( void )
{

  peer_t *res = NULL;


  if (!ring->enabled) return NULL;

  forall_friends(f,
    if (f->from_predecessor && f->status & (OK | TOCALL | PROBABLYDEAD)) {
      res = f;
      break;
    }
  );
  return res;

}



/*
 * successor(): the first peer descriptor in the neighbour list
 *
 * Returns ptr to peer descriptor if found and not "self"
 * (this is a very special situation when a ring is just started),
 * NULL otherwise.
 */
peer_t *
successor
( void )
{

  peer_t *f;


  if (!ring->enabled) return NULL;

  f = ring->neigh_list;
  if (f != NULL && !f->self)
    return f;
  return NULL;

}



/*
 * from_predecessor(): check if a given established incoming connection
 * (a "friend") is marked as "predecessor", that is, it comes from
 * the peer who welcomed the local peer in the ring.
 *
 * Returns ptr to friend if flagged as such, NULL otherwise.
 *
 * Caller must NOT have raised ring->lock
 */
peer_t *
from_predecessor
( csocket_t *cs )
{

  peer_t *res = NULL;


  if (!ring->enabled) return NULL;

  if (cs == NULL) return NULL;

  forall_friends(f,
    if (f->status & (OK | TOCALL | PROBABLYDEAD))
      /*  We are in friend list, so *f->in_cs != NULL for sure.  */
      if ((cs == *f->in_cs) && (f->from_predecessor)) {
        res = f;
        break;
      }
  );
  return res;

}



/*
 * from_successor(): check if a given established incoming connection
 * (a "friend") is marked as "successor", that is, it is the backward
 * connection established by the successor peer.
 *
 * Returns ptr to friend if flagged as such, NULL otherwise.
 *
 * Caller must NOT have raised ring->lock
 */
peer_t *
from_successor
( csocket_t *cs )
{

  peer_t *res = NULL;


  if (!ring->enabled) return NULL;

  if (cs == NULL) return NULL;

  forall_friends(f,
    if (f->status & (OK | TOCALL | PROBABLYDEAD))
      /*  We are in friend list, so *f->in_cs != NULL for sure.  */
      if ((cs == *f->in_cs) && (f->from_successor)) {
        res = f;
        break;
      }
  );
  return res;

}



/*
 * search_friend(): search the friend list by <IPaddr,port> (only living ones),
 * starting from a given point in the list.
 *
 * Returns ptr to friend if found, NULL otherwise.
 *
 * Caller must NOT have raised ring->lock
 */
peer_t *
search_friend
( peer_t *list_head, struct in_addr ipaddr, u_int16_t port )
{

  peer_t *f;


  if (!ring->enabled) return NULL;

  pthread_mutex_lock(&ring->lock);
  for (f=list_head; f!=NULL; f=f->next)
    if (f->status & (OK | TOCALL | PROBABLYDEAD))
      /*  Equivalence between this host's IP addr and "localhost"  */
      if ((ipaddr.s_addr == f->addr.s_addr ||
          (ipaddr.s_addr == my_ipaddr_alt.s_addr &&
          f->addr.s_addr == my_ipaddr.s_addr) ||
          (ipaddr.s_addr == my_ipaddr.s_addr &&
          f->addr.s_addr == my_ipaddr_alt.s_addr))
          && (port == f->port)) {
        pthread_mutex_unlock(&ring->lock);
        return f;
      }
  pthread_mutex_unlock(&ring->lock);
  return NULL;

}



/*
 * empty_neighbour_list(): check if current neighbour list is empty.
 *
 * Returns 1 if so, 0 otherwise.
 *
 * Caller must NOT have raised ring->lock
 */
int
empty_neighbour_list(void)
{

  int           res;


  if (!ring->enabled) return 1;

  pthread_mutex_lock(&ring->lock);
  res = ring->neigh_list == NULL || ring->neigh_list->self;
  pthread_mutex_unlock(&ring->lock);
  return res;

}



/*
 * is_successor(): check if a given peer descriptor is the one corresponding
 * to our nearest neighbour.
 *
 * Returns ptr to peer if requestor was such, NULL otherwise.
 *
 * Caller must NOT have raised ring->lock
 */
peer_t *
is_successor
( peer_t *p )
{

  peer_t *res;


  if (!ring->enabled) return NULL;

  if (p == NULL) return NULL;

  res = NULL;
  /*  Successor is always first in the neigh list so check first one only  */
  forall_neighbours(f,
    if ((p->addr.s_addr == f->addr.s_addr) && (p->port == f->port)) {
      res = f;
      break;
    } else break;
  );
  return res;

}



/*
 * is_neighbour(): check if a given peer identified by <ip_addr,port>
 * is a neighbour, by searching the neighbour list.
 * The entire list is searched, including PROBABLYDEAD entries.
 *
 * Returns ptr to neighbour if requestor was such, NULL otherwise.
 *
 * Caller must NOT have raised ring->lock
 */
peer_t *
is_neighbour
( unsigned long ipaddr, u_int16_t port )
{

  peer_t *res;


  if (!ring->enabled) return NULL;

  res = NULL;
  forall_neighbours(f,
    if (f->status & (OK | TOCALL | PROBABLYDEAD))
      /*  Equivalence between this host's IP addr and "localhost"  */
      if ((ipaddr == f->addr.s_addr ||
          (ipaddr == my_ipaddr_alt.s_addr &&
          f->addr.s_addr == my_ipaddr.s_addr) ||
          (ipaddr == my_ipaddr.s_addr &&
          f->addr.s_addr == my_ipaddr_alt.s_addr))
          && (port == f->port)) {
        res = f;
        break;
      }
  );
  return res;

}



/*
 * is_neighbour_or_finger(): check if a given peer is a neighbour/finger,
 * by searching the neighbour and finger tables.
 *
 * Returns ptr to neighbour/finger if requestor was such, NULL otherwise.
 *
 * Caller must NOT have raised ring->lock
 */
peer_t *
is_neighbour_or_finger
( peer_t *p)
{

  peer_t *res;


  if (!ring->enabled) return NULL;

  if (p == NULL) return NULL;

  res = NULL;
  forall_peers(f,
    if (f->status & (OK | TOCALL | PROBABLYDEAD))
      /*  Equivalence between this host's IP addr and "localhost"  */
      if ((p->addr.s_addr == f->addr.s_addr ||
          (p->addr.s_addr == my_ipaddr_alt.s_addr &&
          f->addr.s_addr == my_ipaddr.s_addr) ||
          (p->addr.s_addr == my_ipaddr.s_addr &&
          f->addr.s_addr == my_ipaddr_alt.s_addr))
          && (p->port == f->port)) {
        res = f;
        break;
      }
  );
  return res;

}



/*
 * Announce to peer "f" that a friendship request is about to arrive to it
 * from peer "foreign_id", whose IP address, port, and pubkey, are given.
 *
 * Returns -2 if "f" is provably dead, -1 on other failures, 0 on success
 */
int
announce_friend
(
  peer_t *f,
  struct in_addr foreign_addr, u_int16_t foreign_port,
  unsigned int predecessor,
  char *foreign_keymod, char *foreign_keyexp,
  pthread_t th_id
)
{

  char *cmd;
  int   rc;


  _do_(_malloc_(cmd,KEYBUFLEN+SLACK),!cmd,"announce_friend(): cmd malloc");

  sprintf(cmd,"announce %lu %u %u %s %s",
          (unsigned long)foreign_addr.s_addr,foreign_port,predecessor,
          foreign_keymod,foreign_keyexp);

  rc = tell_peer(f,is_successor(f) ? 1 : 0,cmd,strlen(cmd)+1,0,th_id);

  if (rc < 0) {
    /* peer F might be down. */
    _free_(cmd);
    return rc;
  }

#ifdef DEBUG_J
  S(fprintf(stderr,"[%lu]\t<< %s (announced friendship)\n",th_id,cmd));
#else
#ifdef DEBUG_F
  S(fprintf(stderr,"[%lu]\t<< %s (announced friendship)\n",th_id,cmd));
#endif
#endif

  _free_(cmd);
  return 0;

}


/* Message routing. */



/*
 *  "Next hop": one more step along a route.  The step is chosen by looking
 *  at remote peers which seem to be alive, and not flagged as "tmpunavail".
 *
 *  Gets:  addr: the destination overlay address.
 *  Returns NULL if addr is stored locally, -1 if don't know where to route,
 *  otherwise ptr to a remote peer.
 *
 *  Caller must not have raised lock over ring struct.
 */
peer_t *
next_hop
( neblo_addr addr )
{

  neblo_addr           dist,d;
  peer_t              *res,*hop;
  unsigned int         i;


  pthread_mutex_lock(&ring->lock);

  if (_belong_(addr,ring->my_id,ring->up_bound)) {
    pthread_mutex_unlock(&ring->lock);
    return NULL;
  }

  dist = _distance_(ring->my_id,addr);
  res = NULL;

  /*
   *  Search among known LIVING peers (neighbours and fingers).
   *  Exclude those peers flagged as "tmpunavail".
   */

  /*  Small distance and default:  search linearly along neighbours  */
  if (successor() != NULL
      && (successor()->status & (OK | TOCALL | PROBABLYDEAD))
      && !(successor()->tmpunavail))
    res = successor();

  if (dist >= CUTOFF)
    /*  Large distance:  search exponentially far away along LIVING fingers */
    /*  Nearest finger is at <= CUTOFF distance  */
    for (i=0, d=CUTOFF; i<NUMFINGERS; i++, d<<=1) {
      if (dist < d)
        break;
      hop = &ring->finger_table[i];
      if ((hop->status & (OK | TOCALL | PROBABLYDEAD)) && !hop->tmpunavail)
        res = hop;
    }

  pthread_mutex_unlock(&ring->lock);

  if (res == NULL)
    /*  Don't know where to route :-(  */
    return (peer_t *)(-1);

  return res;

}



/*
 * forward a msg of kind write (unidirectional), read (request), or
 * data (response) to another peer.
 * msg already provides room for the neblo header.
 * parameter "h" (read handle) is only significant with read and data messages.
 * if flag == 1, status of dest. peer is also marked as "tmpunavail" in case
 * of failure (function tell_peer() accomplishes this).
 */
int
forward_msg
(
  rwd_t kind, peer_t *f, neblo_port port, neblo_addr addr,
  neblo_addr real_addr, char *msg, unsigned int len,
  neblo_rhandle handle, int flag, pthread_t th_id
)
{

  int rc;


#ifdef DEBUG_RW
  S(fprintf(stderr,"[%lu]\tforward_msg(): sending msg to %u.%u.%u.%u:%u (status %u)\n",
            th_id,
            ((unsigned char *)&f->addr.s_addr)[0],
            ((unsigned char *)&f->addr.s_addr)[1],
            ((unsigned char *)&f->addr.s_addr)[2],
            ((unsigned char *)&f->addr.s_addr)[3],
            f->port,f->status));
#endif

  /*
   * attach a header in front of the "msg" payload.
   * to avoid useless memcpy, the caller already provides enough room
   * heading the actual data in the buffer, to hold the header
   */
  switch (kind) {
    case WRITE:
      sprintf(msg,"write %04x %08x %016llx %016llx",
              port,(u_int32_t)len,(ULL)addr,(ULL)real_addr);
      /* replace trailing \0 by " " */
      memset(msg+HEADERSIZE_W-1,32,1);
      /* forward... */
      rc = tell_peer(f,is_successor(f) ? 1 : 0,msg,len+HEADERSIZE_W,flag,
                     th_id);
      break;
    case READ:
      sprintf(msg,"read %04x %08x %08x %016llx %016llx",
              port,(u_int32_t)len,handle,(ULL)addr,(ULL)real_addr);
      /* replace trailing \0 by " ".  Also read msgs can carry a payload! */
      memset(msg+HEADERSIZE_R-1,32,1);
      /* forward... */
      rc = tell_peer(f,is_successor(f) ? 1 : 0,msg,len+HEADERSIZE_R,flag,
                     th_id);
      break;
    case DATA:
      sprintf(msg,"data %04x %08x %08x %016llx %016llx",
              port,(u_int32_t)len,handle,(ULL)addr,(ULL)real_addr);
      /* replace trailing \0 by " " */
      memset(msg+HEADERSIZE_R-1,32,1);
      /* forward... */
      rc = tell_peer(f,is_successor(f) ? 1 : 0,msg,len+HEADERSIZE_R,flag,
                     th_id);
      break;
    case REFR:
      /* not handled here because it needs different parms */
      rc = -1;
      break;
    default:
      rc = -1;
      break;
 }

#ifdef DEBUG_RW
  if (rc >= 0)
    S(fprintf(stderr,"[%lu]\t<< %s\n",th_id,msg));
#endif

  return rc;

}



/*
 * invoke the application-level deliver handler for write msgs
 * at the destination peer
 */
int
invoke_receiver_handler_UD
( neblo_port port, neblo_addr addr, void *msg, unsigned int len )
{
  int rc = 0;
  if (__receiver_handler_UD[port] != NULL) {
    pthread_mutex_lock(&__receiver_handler_lock);
    rc = __receiver_handler_UD[port](port,addr,msg,&len);
    pthread_mutex_unlock(&__receiver_handler_lock);
  }
  return rc;
}



/*
 * invoke the application-level deliver handler for read msgs at the
 * destination peer.
 * the handler can change the content of region "msg", and set "actual_len"
 * to a different (smaller) value.
 */
int
invoke_receiver_handler_REQ
( neblo_port port, neblo_addr addr, void *msg, unsigned int *actual_len )
{
  int rc = 0;
  if (__receiver_handler_REQ[port] != NULL) {
    pthread_mutex_lock(&__receiver_handler_lock);
    rc = __receiver_handler_REQ[port](port,addr,msg,actual_len);
    pthread_mutex_unlock(&__receiver_handler_lock);
  }
  return rc;
}



/*
 * invoke the application-level forward handler for write msgs
 * at each intermediate peer, including the initial one (but not the
 * destination).
 * the handler will return -1 if msg is not to be forwarded by the neblo
 * layer; otherwise returns 0
 */
int
invoke_forwarder_handler_UD
( neblo_port port, neblo_addr addr, void *msg, unsigned int len )
{
  int rc = 0;
  if (__forwarder_handler_UD[port] != NULL) {
    pthread_mutex_lock(&__forwarder_handler_lock);
    rc = __forwarder_handler_UD[port](port,addr,msg,&len);
    pthread_mutex_unlock(&__forwarder_handler_lock);
  }
  return rc;
}



/*
 * invoke the application-level forward handler for read msgs
 * at each intermediate peer, including the initial one (but not the
 * destination).
 * the handler will return -1 if msg is not to be forwarded by the neblo
 * layer; otherwise returns 0
 */
int
invoke_forwarder_handler_REQ
( neblo_port port, neblo_addr addr, void *msg, unsigned int maxlen )
{
  int rc = 0;
  if (__forwarder_handler_REQ[port] != NULL) {
    pthread_mutex_lock(&__forwarder_handler_lock);
    rc = __forwarder_handler_REQ[port](port,addr,msg,&maxlen);
    pthread_mutex_unlock(&__forwarder_handler_lock);
  }
  return rc;
}



/*
 * invoke the application-level forward handler for "data" msgs
 * at each intermediate peer of the trip back, NOT including the initial
 * as well as the final one.
 * will always return 0
 */
int
invoke_forwarder_handler_RES
( neblo_port port, neblo_addr addr, void *msg, unsigned int len )
{
  int rc = 0;
  if (__forwarder_handler_RES[port] != NULL) {
    pthread_mutex_lock(&__forwarder_handler_lock);
    rc = __forwarder_handler_RES[port](port,addr,msg,&len);
    pthread_mutex_unlock(&__forwarder_handler_lock);
  }
  return rc;
}



/*
 * invoke the application-level handler to re-publish data associated to
 * the given addr interval, so as to re-create replicas
 */
int
invoke_refresh_handler
( neblo_addr lo_addr, neblo_addr up_addr, neblo_addr req_addr )
{
  int rc = 0;
  if (__refresh_handler != NULL) {
    pthread_mutex_lock(&__receiver_handler_lock);
    __refresh_in_progress = 1;
    __refresh_addr = req_addr;
    rc = __refresh_handler(lo_addr,up_addr);
    __refresh_in_progress = 0;
    pthread_mutex_unlock(&__receiver_handler_lock);
  }
  return rc;
}



/*
 * invoke the application-level handler to cancel blocks belonging to
 * an addr interval no longer mapped to the local peer
 */
int
invoke_clean_handler
( neblo_addr lo_addr, neblo_addr up_addr )
{
  int rc = 0;
  if (__clean_handler != NULL) {
    pthread_mutex_lock(&__receiver_handler_lock);
    rc = __clean_handler(lo_addr,up_addr);
    pthread_mutex_unlock(&__receiver_handler_lock);
  }
  return rc;
}


/* Joining the ring. */



/*
 *  Fold a "long" word (src) over a shorter one (dest) by xor.
 *  Error if src is actually shorter than dest.
 */
void
xor_fold
( int8_t *dest, unsigned int dest_size, int8_t *src, unsigned int src_size )
{

  unsigned int  ds;
  int8_t       *d;


  if (src_size < dest_size) {
    fprintf(stderr,"xor_fold():  error: source shorter than dest\n");
    return;
  }
  memcpy(dest,src,dest_size);
  if (src_size == dest_size) {
    /*  No need to fold  */
    return;
  }
  /*  Fold over in chunks, each dest_size long.  First chunk already there.  */
  src += dest_size;
  src_size -= dest_size;
  d = dest;
  ds = dest_size;
  for (; src_size; src_size--, src++) {
    *d ^= *src;
    if (!--ds) {
      d = dest;
      ds = dest_size;
    } else
      d++;
  }
  return;

}


/*  Handles for recursive-style commands.  */
/*  Some of them are macros, see neblo_.h  */



void
cmd_handle_init
( cmd_handle_t *h )
{
  h->val = 0;
  _pthmutexinit_(&h->lock);
  return;
}



void
cmd_handle_set
( cmd_handle_t *h )
{
  RAND_bytes((unsigned char *)&h->val,sizeof(h->val));
  gettimeofday(&h->timestamp,NULL);
  return;
}



int
cmd_handle_age
( cmd_handle_t *h )
{
  struct timeval now;

  gettimeofday(&now,NULL);
  return (int)now.tv_sec - (int)h->timestamp.tv_sec;
}



/*
other cmd_handle routines, defined as macros in neblo_.h:

cmd_handle_reset(h) (h)->val = 0;
cmd_handle_val(h) ((h)->val)
cmd_handle_lock(h) pthread_mutex_lock(&(h)->lock);
cmd_handle_unlock(h) pthread_mutex_unlock(&(h)->lock);
cmd_handle_is_done(h) ((h)->val == 0)
*/
