/*********************************************************************/
/* newfinger.c                                                       */
/* function neblo_newfinger(), manages responses to "findfinger" cmd */
/* see COPYING in parent directory                                   */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>


int 
neblo_newfinger
( reqst_t *reqst )
{
  
  u_int32_t             f_handle_val;
  unsigned int          magn;
  unsigned int          fport;
  u_int16_t             f_port;
  char                 *keymod;
  char                 *keyexp;
  unsigned int          fidx;
  peer_t               *f;
  peer_t               *ff;
#ifdef DEBUG_R
  unsigned int          i;
#endif
  int                   rc;


  _do_(_malloc_(keymod,KEYBUFLEN+SLACK),
       !keymod,"neblo_newfinger(): keymod malloc");
  _do_(_malloc_(keyexp,BUFLEN+SLACK),
       !keyexp,"neblo_newfinger(): keyexp malloc");

  sscanf(reqst->buf,"%*s %u %u %u %50000s %100s",
         &f_handle_val,&magn,&fport,keymod,keyexp);
  f_port = (u_int16_t)fport;

  /*
   *  SECURITY:  accept only if "newfinger" carries the 32 bit handle
   *  originally associated to the "findfinger" req.  That handle can
   *  only be associated to a free entry in the finger table, as this
   *  represents the missing finger.
   */

  for (fidx=0; fidx<NUMFINGERS; fidx++) {
    f = &ring->finger_table[fidx];
    cmd_handle_lock(&f->handle);
    if (f_handle_val == cmd_handle_val(&f->handle))
      break;
    cmd_handle_unlock(&f->handle);
  }

  if (fidx >= NUMFINGERS) {

#ifdef DEBUG_R
    S(fprintf(stderr,"[%lu]\thandle not found. discard.\n",reqst->th_id));
#endif

    goto abort1;
  }

  /*  SANITY: Index "fidx" should amount to "magn".  */

  if (fidx != magn) {

#ifdef DEBUG_R
    S(fprintf(stderr,"[%lu]\twrong magnitude. discard.\n",reqst->th_id));
#endif

    goto abort1;
  }

  /*  New finger put a challenge on our pubkey.  */

  rc = _get_challenge_(reqst->csock,reqst->buf);

  if (rc < 0) {

#ifdef DEBUG_R
    S(fprintf(stderr,"[%lu]\tchallenge interrupted\n",reqst->th_id));
#endif

    goto abort2;
  }

#ifdef DEBUG_R
  S(fprintf(stderr,"[%lu]\tchallenge done\n",reqst->th_id));
#endif

  /*  Prepare to establish a friendship with new finger.  */

  clean_peer_desc_nolock(f);
  f->addr.s_addr = reqst->peer_addr.s_addr;
  f->port = f_port;
  _do_(_RSA_new_(f->pubkey),!f->pubkey,
       "neblo_newfinger(): RSA_new of f->pubkey");

  _convert_pubkey_(f->pubkey,keymod,keyexp,goto abort);

  f->status = TOCALL;

  rc = call_peer(f,0,0,reqst->th_id);

  if (rc < 0) {

#ifdef DEBUG_R
    S(fprintf(stderr,"[%lu]\tFailed friendship with new finger\n",
              reqst->th_id));
#endif

    goto abort;
  }

  _write_(reqst->csock,"ok",strlen("ok")+1,rc,goto abort);

#ifdef DEBUG
  S(fprintf(stderr,"[%lu]\t<< ok\n",reqst->th_id));
#endif

#ifdef DEBUG_R
  S(
  fprintf(stderr,"[%lu]\tAdded a new finger.  New finger table:\n",
          reqst->th_id);
  pthread_mutex_lock(&ring->lock);
  for (i=0; i<NUMFINGERS; i++) {
    ff = &ring->finger_table[i];
    if (ff->status & (OK | TOCALL | PROBABLYDEAD))
      fprintf(stderr,"\tmagnitude %u (presumable dist. %016llx):  %u.%u.%u.%u:%u (%u)\n",
                i,finger_distance_rule(i),
                ((unsigned char *)&ff->addr.s_addr)[0],
                ((unsigned char *)&ff->addr.s_addr)[1],
                ((unsigned char *)&ff->addr.s_addr)[2],
                ((unsigned char *)&ff->addr.s_addr)[3],
                ff->port,ff->status);
  }
  pthread_mutex_unlock(&ring->lock);
  );
#endif

  rc = 0;

  cmd_handle_unlock(&f->handle);


  if (magn < NUMFINGERS-1) {

    /*
     *  Find finger F(magn+1), based on the new finger F(magn) we have just
     *  contacted.  Although such a finger might already exist in the local
     *  finger table, we prefer to recompute it as it might be inconsistent
     *  with the new one in terms of distance rule.
     */

    ff = &ring->finger_table[++magn];

#ifdef DEBUG_R
    S(fprintf(stderr,"[%lu]\tissuing findfinger req for magnitude %u\n",
              reqst->th_id,magn));
#endif

    /*  Be sure no other thread is doing the same (should not occur).  */

    cmd_handle_lock(&ff->handle);

#if 0
    if (!cmd_handle_is_done(&ff->handle)) {
      if (cmd_handle_age(&ff->handle) <= (int)REC_CMD_DEADLINE) {

#ifdef DEBUG_R
        S(fprintf(stderr,
                  "[%lu]\tfindfinger cmd still in progress from last time\n",
                  reqst->th_id));
#endif

      } else
        cmd_handle_reset(&ff->handle);
    }
#endif

    if (cmd_handle_is_done(&ff->handle)) {

      /*  Create handle for this recursive request.  */
      cmd_handle_set(&ff->handle);

      sprintf(reqst->buf,"findfinger %u %u %016llx %lu %u %u %s %s",
              magn,
              0              /* scale factor is 0 when magn > 0           */,
              (ULL)(neblo_addr)1  /* addr is 1 when request has magn > 0  */,
              (unsigned long)my_ipaddr.s_addr,PORT,
              cmd_handle_val(&ff->handle),
              pubkeymod,pubkeyexp);

      rc = tell_peer(f,0,reqst->buf,strlen(reqst->buf)+1,0,reqst->th_id);

      if (rc < 0) {
        /*  Failed to emit "findfinger" req.  Try later.  */
        cmd_handle_reset(&ff->handle);
      } else {

#ifdef DEBUG_R
        S(fprintf(stderr,"[%lu]\t<< %s\n",reqst->th_id,reqst->buf));
#endif

      }

    }

    cmd_handle_unlock(&ff->handle);

  }

exit:

  _close_(&reqst->csock);
  _free_(keymod); _free_(keyexp);
  return rc;

abort:

  _RSA_free_(f->pubkey);
  f->status = DEAD;

abort2:

  cmd_handle_unlock(&f->handle);

abort1:

  rc = -1;
  goto exit;  

}
